//*****************************************************************************
//
// main.c - The microcontroller hardware configuration and SafeRTOS initializing
//          for the ACC PIN Reader application.
//
// Code of this module is based on the qs-safertos example that is provided as
// a part of revision 5585 of the EK-LM3S9B96 Firmware Package.
//
//*****************************************************************************

#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "driverlib/gpio.h"
#include "driverlib/rom.h"
#include "driverlib/sysctl.h"
#include "utils/uartstdio.h"
#include "SafeRTOS/SafeRTOS_API.h"
#include "led_task.h"
#include "uart_task.h"
#include "display_task.h"
#include "WGNDin_task.h"
#include "WGNDout_task.h"
#include "touch_task.h"
#include "reader_task.h"

//*****************************************************************************
//
// This application utilizes SafeRTOS to perform a variety of tasks in a
// concurrent fashion. The following tasks are created:
//
// * A reader task that integrates into one coherent program all other tasks and
//   controls upper level functionality of the ACC PIN Reader.
//
// * A display task that services the keypad pattern display.
//
// * A touch panel task that services the touch panel.
//
// * A Wiegand inputs and indicators task that controls inputs of the Wiegand
//   interface and sets the ACC PIN Reader indicators accordingly.
//
// * A Wiegand outputs task that transmits via Wiegand interface output lines
//   Data0 and Data1 code of the pressed key.
//
// * An UART task that sends to the user via the UART messages from SafeRTOS
//   kernel and other performed tasks as well as provides some basic information
//   at startup.
//
// * A LED task that blinks on the EKK-LM3S9B96 board the user LED.
//
// There is an automatically created idle task, which does nothing in this project.
//
//*****************************************************************************

//*****************************************************************************
//
// The error routine that is called if the driver library encounters an error.
//
//*****************************************************************************
#ifdef DEBUG
void
__error__(char *pcFilename, unsigned long ulLine)
{
}
#endif

//*****************************************************************************
//
// This hook is called by SafeRTOS when a task is deleted.
//
//*****************************************************************************
void
SafeRTOSTaskDeleteHook(xTaskHandle xTaskToDelete)
{
}

//*****************************************************************************
//
// The stack for the idle task.
//
//*****************************************************************************
unsigned long g_pulIdleTaskStack[128];

//*****************************************************************************
//
// This hook is called by the SafeRTOS idle task when no other tasks are
// runnable.
//
//*****************************************************************************
void
SafeRTOSIdleHook(void)
{
}

//*****************************************************************************
//
// This hook is called by SafeRTOS when an error is detected.
//
//*****************************************************************************
static void
SafeRTOSErrorHook(xTaskHandle xHandleOfTaskWithError,
                  signed portCHAR *pcNameOfTaskWithError,
                  portBASE_TYPE xErrorCode)
{
    //
    // A fatal SafeRTOS error was detected, so output an error message.
    //
    UARTprintf("Fatal SafeRTOS error!\n");

    //
    // This function can not return, so loop forever.  Interrupts are disabled
    // on entry to this function, so no processor interrupts will interrupt
    // this loop.
    //
    while(1)
    {
    }
}

//*****************************************************************************
//
// The parameters used to initialize SafeRTOS.
//
//*****************************************************************************
static xPORT_INIT_PARAMETERS g_sSafeRTOSPortInit =
{
    //
    // System clock rate.
    //
    80000000,

    //
    // Scheduler tick rate.
    //
    1000 / portTICK_RATE_MS,

    //
    // Task delete hook.
    //
    SafeRTOSTaskDeleteHook,

    //
    // Error hook.
    //
    SafeRTOSErrorHook,

    //
    // Idle hook.
    //
    SafeRTOSIdleHook,

    //
    // System stack location.
    //
    0,

    //
    // System stack size.
    //
    0,

    //
    // Vector table base.
    //
    0
};

//*****************************************************************************
//
// Initialize SafeRTOS and start the initial set of tasks.
//
//*****************************************************************************
int
main(void)
{
    //
    // Set the clocking to run at 80 MHz from the PLL.
    //
    ROM_SysCtlClockSet(SYSCTL_SYSDIV_2_5 | SYSCTL_USE_PLL | SYSCTL_XTAL_16MHZ |
                       SYSCTL_OSC_MAIN);

    //
    // Initialize the UART output.
    //
    ROM_SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOA);
    ROM_GPIOPinTypeUART(GPIO_PORTA_BASE, GPIO_PIN_0 | GPIO_PIN_1);
    UARTStdioInit(0);

    //
    // Print a welcome banner.
    //
    UARTprintf("Random Keypad Pattern ACC PIN Reader v.1.0\n");

    //
    // Set the location and size of the system stack.
    //
    g_sSafeRTOSPortInit.pulSystemStackLocation =
        (unsigned portLONG *)(*(unsigned long *)0);
    g_sSafeRTOSPortInit.ulSystemStackSizeBytes = 128 * 4;

    //
    // Initialize the SafeRTOS kernel.
    //
    vTaskInitializeScheduler((signed portCHAR *)g_pulIdleTaskStack,
                             sizeof(g_pulIdleTaskStack), 0,
                             &g_sSafeRTOSPortInit);
    
	//
    // Create the Wiegand outputs service task
    //
    if(WGNDoutTaskInit() != 0)
    {
        UARTprintf("Failed to create WGNDout task!\n");
        while(1)
        {
        }
    }

	//
    // Create the Wiegand inputs and reader indicators service task
    //
    if(WGNDinTaskInit() != 0)
    {
        UARTprintf("Failed to create WGNDin task!\n");
        while(1)
        {
        }
    }
	
    //
    // Create the keypad display service task
    //
    if(DisplayTaskInit() != 0)
    {
        UARTprintf("Failed to create Display task!\n");
        while(1)
        {
        }
    }

	//
    // Create the touch panel service task
    //
    if(TouchPanelTaskInit() != 0)
    {
        UARTprintf("Failed to create Touch Panel task!\n");
        while(1)
        {
        }
    }

    //
    // Create the reader task
    //
    if(ReaderTaskInit() != 0)
    {
        UARTprintf("Failed to create Reader main task!\n");
        while(1)
        {
        }
    }

	//
    // Create the LED task
    //
    if(LEDTaskInit() != 0)
    {
        UARTprintf("Failed to create LED task!\n");
        while(1)
        {
        }
    }

    //
    // Create the UART task
    //
    if(UARTTaskInit() != 0)
    {
        UARTprintf("Failed to create UART task!\n");
        while(1)
        {
        }
    }

    //
    // Start the scheduler.  This should not return
    //
    xTaskStartScheduler(pdTRUE);

    //
    // In case the scheduler returns for some reason, print an error and loop
    // forever.
    //
    UARTprintf("Failed to start scheduler!\n");
    while(1)
    {
    }
}
