//*****************************************************************************
//
// WGNDin_task.c - Wiegand inputs and reader indicators service task
//
//*****************************************************************************

#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "inc/hw_gpio.h"
#include "driverlib/gpio.h"
#include "driverlib/rom.h"
#include "driverlib/sysctl.h"
#include "SafeRTOS/SafeRTOS_API.h"
#include "WGNDin_task.h"
#include "priorities.h"

//*****************************************************************************
//
// The GPIO pins to which Wiegand inputs and reader indicators are connected
//
//*****************************************************************************
// Wiegand inputs
#define WGND_IN_P_PERIPH     SYSCTL_PERIPH_GPIOB
#define WGND_IN_P_BASE       GPIO_PORTB_BASE
#define WGND_IN_LED_R        GPIO_PIN_5
#define WGND_IN_LED_G        GPIO_PIN_6
#define WGND_IN_BUZ          GPIO_PIN_7
// Reader indicators
#define WGND_IND_P_PERIPH    SYSCTL_PERIPH_GPIOG
#define WGND_IND_P_BASE      GPIO_PORTG_BASE
#define WGND_LED_R           GPIO_PIN_0
#define WGND_LED_G           GPIO_PIN_1
#define WGND_BUZ             GPIO_PIN_7

//*****************************************************************************
//
// The amount of time to delay between successive reads of Wiegand inputs
//
//*****************************************************************************
const portTickType g_xWGNDinReadDelay = 10;

//*****************************************************************************
//
// The LED_Red input: the input state, input filter and input stable condition
//
//*****************************************************************************
static unsigned char g_ucInLEDRedState;
static unsigned char g_ucInLEDRedFilter;
#define WGND_IN_LED_R_STABLE	3

//*****************************************************************************
//
// The LED_Green input: the input state, input filter and input stable condition
//
//*****************************************************************************
static unsigned char g_ucInLEDGreenState;
static unsigned char g_ucInLEDGreenFilter;
#define WGND_IN_LED_G_STABLE	3

//*****************************************************************************
//
// The Buzzer input: the input state, input filter and input stable condition
//
//*****************************************************************************
static unsigned char g_ucInBuzzerState;
static unsigned char g_ucInBuzzerFilter;
#define WGND_IN_BUZ_STABLE		3

//*****************************************************************************
//
// The timer controlling beep sound
//
//*****************************************************************************
static unsigned char g_ucBeepTimer = 0;

//*****************************************************************************
//
// The stack for the WGNDin task
//
//*****************************************************************************
static unsigned long g_pulWGNDinTaskStack[128];

//*****************************************************************************
//
// The item size, queue size, and memory size for the beeper input queue
//
//*****************************************************************************
#define BUZZER_ITEM_SIZE		sizeof(g_ucBeepTimer)
#define BUZZER_QUEUE_LENGTH		1
#define BUZZER_QUEUE_MEM_SIZE	((BUZZER_ITEM_SIZE * BUZZER_QUEUE_LENGTH) + \
                                 portQUEUE_OVERHEAD_BYTES)

//*****************************************************************************
//
// A buffer to contain the contents of the beeper input queue
//
//*****************************************************************************
signed portCHAR g_cBuzzerQueueBuffer[BUZZER_QUEUE_MEM_SIZE];

//*****************************************************************************
//
// The queue that holds messages sent to the buzzer
//
//*****************************************************************************
static xQueueHandle g_pBuzzerQueue;

//*****************************************************************************
//
// Maximum delay to make beep
//
//*****************************************************************************
#define BEEP_MAX_DELAY	11

//*****************************************************************************
//
// Sends a request to the buzzer to make one beep
//
//*****************************************************************************
void
MakeBeep(unsigned char ucBeepTime)
{
    //
    // Send the beep request to the buzzer
    //
    xQueueSend(g_pBuzzerQueue, &ucBeepTime, BEEP_MAX_DELAY);
}

//*****************************************************************************
//
// This task reads Wiegand inputs: LED_Red, LED_Green, BUZzer and activates
// reader indicators: red LED, green LED and Buzzer accordingly. The indicator
// is turned on if its coorresponding input is pulled LOW.
// The task implements simple filter for each input that cuts out input glitches.
// The input is assumed to be stable if its state does not change during three
// consecutive reads.
//
//*****************************************************************************
static void
WGNDinTask(void *pvParameters)
{
    portTickType ulLastTime;
	unsigned char ucInState;

    //
    // Get the current tick count.
    //
    ulLastTime = xTaskGetTickCount();

    //
    // Loop forever.
    //
    while(1)
    {
		//
		// Read the Wiegand LED_Red input
		//
		ucInState = (ROM_GPIOPinRead(WGND_IN_P_BASE, WGND_IN_LED_R));
		//
		// Filter the LED_Red input state
		//
		if (ucInState != g_ucInLEDRedState){
		  //
		  // The input state has changed - reset input filter
		  //
		  g_ucInLEDRedState = ucInState;
		  g_ucInLEDRedFilter = 0;
		}
		else {
		  //
		  // The input state hasn't changed - check input filter
		  //
		  if (g_ucInLEDRedFilter < WGND_IN_LED_R_STABLE){
		    //
			// Input filter active - service it
			//
			g_ucInLEDRedFilter++;
			if (g_ucInLEDRedFilter >= WGND_IN_LED_R_STABLE){
			  //
			  // The input state has stabilized - update the red LED
			  //
			  ROM_GPIOPinWrite(WGND_IND_P_BASE, WGND_LED_R, (g_ucInLEDRedState ? 0 : WGND_LED_R));
			}
		  }
		}

		//
		// Read the Wiegand LED_Green input
		//
		ucInState = (ROM_GPIOPinRead(WGND_IN_P_BASE, WGND_IN_LED_G));
		//
		// Filter the LED_Green input state
		//
		if (ucInState != g_ucInLEDGreenState){
		  //
		  // The input state has changed - reset input filter
		  //
		  g_ucInLEDGreenState = ucInState;
		  g_ucInLEDGreenFilter = 0;
		}
		else {
		  //
		  // The input state hasn't changed - check input filter
		  //
		  if (g_ucInLEDGreenFilter < WGND_IN_LED_G_STABLE){
		    //
			// Input filter active - service it
			//
			g_ucInLEDGreenFilter++;
			if (g_ucInLEDGreenFilter >= WGND_IN_LED_G_STABLE){
			  //
			  // The input state has stabilized - update the green LED
			  //
			  ROM_GPIOPinWrite(WGND_IND_P_BASE, WGND_LED_G, (g_ucInLEDGreenState ? 0 : WGND_LED_G));
			}
		  }
		}

		//
		// Check the buzzer input queue
		//
		if(xQueueReceive(g_pBuzzerQueue, &g_ucBeepTimer, 0) == pdPASS){
		  //
		  // Request for beep - turn on the buzzer
		  //
		  ROM_GPIOPinWrite(WGND_IND_P_BASE, WGND_BUZ, WGND_BUZ);
		}
		
		if (g_ucBeepTimer != 0){
		  //
		  // Request for beep - service beep timer ignoring the Wiegand Buzzer input
		  //
		  g_ucBeepTimer--;
		  if (g_ucBeepTimer == 0){
		    //
			// End of beep - set the buzzer	to its pre-beep request state
			//
			ROM_GPIOPinWrite(WGND_IND_P_BASE, WGND_BUZ, (g_ucInBuzzerState ? 0 : WGND_BUZ));
		  }
		}
		else{
		  //
		  // No request for beep - read the Wiegand Buzzer input
		  //
		  ucInState = (ROM_GPIOPinRead(WGND_IN_P_BASE, WGND_IN_BUZ));
		  //
		  // Filter the Buzzer input state
		  //
		  if (ucInState != g_ucInBuzzerState){
		    //
		    // The input state has changed - reset input filter
		    //
		    g_ucInBuzzerState = ucInState;
		    g_ucInBuzzerFilter = 0;
		  }
		  else {
		    //
		    // The input state hasn't changed - check input filter
		    //
		    if (g_ucInBuzzerFilter < WGND_IN_BUZ_STABLE){
		      //
			  // Input filter active - service it
			  //
			  g_ucInBuzzerFilter++;
			  if (g_ucInBuzzerFilter >= WGND_IN_BUZ_STABLE){
			    //
			    // The input state has stabilized - update the buzzer
			    //
			    ROM_GPIOPinWrite(WGND_IND_P_BASE, WGND_BUZ, (g_ucInBuzzerState ? 0 : WGND_BUZ));
			  }
		    }
		  }
		}

        //
        // Wait for the required amount of time.
        //
        xTaskDelayUntil(&ulLastTime, g_xWGNDinReadDelay);
    }
}

//*****************************************************************************
//
// Initializes Wiegand inputs and reader indicators service task
//
//*****************************************************************************
unsigned long
WGNDinTaskInit(void)
{
    //
    // Initialize the GPIO used as Wiegand inputs
	// PB7 needs unlocking prior to its configuration ! 
    //
    ROM_SysCtlPeripheralEnable(WGND_IN_P_PERIPH);

#if (WGND_IN_P_BASE == GPIO_PORTB_BASE)							// if port PB is used
	HWREG(WGND_IN_P_BASE + GPIO_O_LOCK) = GPIO_LOCK_KEY_DD;		// unlock GPIO_CR register
	HWREG(WGND_IN_P_BASE + GPIO_O_CR) |= GPIO_PIN_7;			// enable configuration of PB7 pin
#endif

    ROM_GPIOPinTypeGPIOInput(WGND_IN_P_BASE,
	                         WGND_IN_LED_R | WGND_IN_LED_G | WGND_IN_BUZ);

#if (WGND_IN_P_BASE == GPIO_PORTB_BASE)							// if port PB is used
	HWREG(WGND_IN_P_BASE + GPIO_O_CR) &= ~(GPIO_PIN_7);			// disable configuration of PB7 pin
	HWREG(WGND_IN_P_BASE + GPIO_O_LOCK) = GPIO_LOCK_M;			// lock GPIO_CR register
#endif

    //
    // Initialize the GPIO used to drive the reader indicators (4mA drive)
    //
    ROM_SysCtlPeripheralEnable(WGND_IND_P_PERIPH);
    ROM_GPIODirModeSet(WGND_IND_P_BASE,
	                   WGND_LED_R | WGND_LED_G | WGND_BUZ,
					   GPIO_DIR_MODE_OUT);
	ROM_GPIOPadConfigSet(WGND_IND_P_BASE,
	                     WGND_LED_R | WGND_LED_G | WGND_BUZ,
						 GPIO_STRENGTH_4MA,
						 GPIO_PIN_TYPE_STD);

    //
    // Create a queue for sending messages to the buzzer
    //
    if(xQueueCreate(g_cBuzzerQueueBuffer, BUZZER_QUEUE_MEM_SIZE,
                    BUZZER_QUEUE_LENGTH, BUZZER_ITEM_SIZE, &g_pBuzzerQueue) != pdPASS)
    {
        return(1);
    }

    //
    // Create the Wiegand input and reader indicators service task
    //
    if(xTaskCreate(WGNDinTask, (signed portCHAR *)"WGNDin",
                   (signed portCHAR *)g_pulWGNDinTaskStack,
                   sizeof(g_pulWGNDinTaskStack), NULL, PRIORITY_WGNDin_TASK,
                   NULL) != pdPASS)
    {
        return(1);
    }

    //
    // Success.
    //
    return(0);
}
