//*****************************************************************************
//
// WGNDout_task.c - Wiegand outputs service task
//
//*****************************************************************************

#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "inc/hw_ints.h"
#include "inc/hw_gpio.h"
#include "inc/hw_timer.h"
#include "driverlib/gpio.h"
#include "driverlib/rom.h"
#include "driverlib/sysctl.h"
#include "driverlib/timer.h"
#include "SafeRTOS/SafeRTOS_API.h"
#include "WGNDout_task.h"
#include "priorities.h"

//*****************************************************************************
//
// The GPIO pins to which Wiegand outputs are connected
//
//*****************************************************************************
#define WGND_OUT_P_PERIPH    SYSCTL_PERIPH_GPIOB
#define WGND_OUT_P_BASE      GPIO_PORTB_BASE
#define WGND_OUT_DATA0       GPIO_PIN_2
#define WGND_OUT_DATA1       GPIO_PIN_3

//*****************************************************************************
//
// The timer which counts duration of the data bit at Wiegand output 
//
//*****************************************************************************
#define WGND_OUT_TIM_PERIPH	SYSCTL_PERIPH_TIMER0
#define WGND_OUT_TIM_BASE	TIMER0_BASE
#define WGND_TIMER			TIMER_A
#define WGND_OUT_PULSE		50
#define WGND_TIMER_INT		INT_TIMER0A
#define WGND_TIMER_TMEOUT	TIMER_TIMA_TIMEOUT

//*****************************************************************************
//
// The amount of time to delay between successive sent bits
//
//*****************************************************************************
const portTickType g_xWGNDoutBitDelay = 2;

//*****************************************************************************
//
// The stack for the WGNDout task
//
//*****************************************************************************
static unsigned long g_pulWGNDoutTaskStack[128];

//*****************************************************************************
//
// The transmitted key code
//
//*****************************************************************************
unsigned char g_ucTxKeyCode;

//*****************************************************************************
//
// The transmission bit counter states, bit masks and the transmission bit counter
//
//*****************************************************************************
#define WGNDOUT_IDLE	0
#define WGNDOUT_TxEVEN	0
#define WGNDOUT_TxBIT3	1
#define	WGNDOUT_TxBIT2	2
#define WGNDOUT_TxBIT1	3
#define WGNDOUT_TxBIT0	4
#define WGNDOUT_TxODD	5
#define WGNDOUT_TxEND	105

#define M_OUT_BIT3		0x08
#define M_OUT_BIT2		0x04
#define M_OUT_BIT1		0x02
#define M_OUT_BIT0		0x01

unsigned char g_ucOutBitCounter = WGNDOUT_IDLE;

//*****************************************************************************
//
// The item size, queue size, and memory size for the WiegandOut task input queue
//
//*****************************************************************************
#define WGNDOUT_ITEM_SIZE		sizeof(g_ucTxKeyCode)
#define WGNDOUT_QUEUE_LENGTH	8
#define WGNDOUT_QUEUE_MEM_SIZE	((WGNDOUT_ITEM_SIZE * WGNDOUT_QUEUE_LENGTH) + \
                                 portQUEUE_OVERHEAD_BYTES)

//*****************************************************************************
//
// A buffer to contain the contents of the WiegandOut input queue
//
//*****************************************************************************
signed portCHAR g_cWGNDoutQueueBuffer[WGNDOUT_QUEUE_MEM_SIZE];

//*****************************************************************************
//
// The queue that holds messages sent to the WGNDout task
//
//*****************************************************************************
static xQueueHandle g_pWGNDoutQueue;

//*****************************************************************************
//
// Maximum delay to send data
//
//*****************************************************************************
#define OUT_MAX_DELAY	2

//*****************************************************************************
//
// Sends a request to the WGNDout task to transmit the key code to the ACC system 
//
//*****************************************************************************
void
WGNDOutputKeyCode(unsigned char ucKeyCode)
{
    //
    // Send the key code to the WGNDout task
    //
    xQueueSend(g_pWGNDoutQueue, &ucKeyCode, OUT_MAX_DELAY);
}

//*****************************************************************************
//
// This task sends via Wiegand outputs: Data0 and Data1 code of the pressed key.
// The data transmission format is compatible with HID ProxPro 5355A reader, i.e.
// each key code is sent as 6-bit long data frame:
// 					E b3 b2
// 							b1 b0 O
// where: E - even parity bit for bits b3 b2
//        b3 b2 b1 b0 - the key code
//         0  0  0  0 : 0
//         0  0  0  1 : 1
//         0  0  1  0 : 2
//         0  0  1  1 : 3
//         0  1  0  0 : 4
//         0  1  0  1 : 5
//         0  1  1  0 : 6
//         0  1  1  1 : 7
//         1  0  0  0 : 8
//         1  0  0  1 : 9
//         1  0  1  0 : *
//         1  0  1  1 : #
//        O - odd parity bit for bits b1 b0
//
//*****************************************************************************
static void
WGNDoutTask(void *pvParameters)
{
    portTickType ulLastTime;
 
    //
    // Get the current tick count
    //
    ulLastTime = xTaskGetTickCount();

    //
    // Loop forever.
    //
    while(1)
    {
		//
		// Check the transmission bit counter
		//
		switch (g_ucOutBitCounter){
		  case WGNDOUT_IDLE		: //
		                          // No transmission in progress - check the input queue
								  //
		                          if(xQueueReceive(g_pWGNDoutQueue, &g_ucTxKeyCode, 0) == pdPASS){
								    //
									// Output even parity bit for the key kode
									//
									if (((g_ucTxKeyCode & M_OUT_BIT3)>>1) != (g_ucTxKeyCode & M_OUT_BIT2)){
									  //
									  // b3 != b2 therefore even parity bit = 1 - pulse Data1 line
									  //
									  ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA1, WGND_OUT_DATA1);
									}
									else{
									  //
									  // b3 == b2 therefore even parity bit = 0 - pulse Data0 line
									  //
									  ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA0, WGND_OUT_DATA0);
									}
									//
		                            // Start counting duration of Data0/Data1 line bit pulse
		                            //
		                            ROM_TimerEnable(WGND_OUT_TIM_BASE, WGND_TIMER);
									//
		                            // Count sent bits
		                            //
								    g_ucOutBitCounter++;
								  }
		  						  break;
								   
		  case WGNDOUT_TxBIT3	: //
		                          // Transmit bit b3
								  //
								  if (g_ucTxKeyCode & M_OUT_BIT3){
								  	//
									// b3 == 1 - pulse Data1 line
								    //
									ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA1, WGND_OUT_DATA1);
								  }
								  else {
								    //
									// b3 == 0 - pulse Data0 line
								    //
									ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA0, WGND_OUT_DATA0);
								  }
								  //
		                          // Start counting duration of Data0/Data1 line bit pulse
		                          //
		                          ROM_TimerEnable(WGND_OUT_TIM_BASE, WGND_TIMER);
								  //
		                          // Count sent bits
		                          //
								  g_ucOutBitCounter++;
		  						  break;

		  case WGNDOUT_TxBIT2	: //
		                          // Transmit bit b2
								  //
		  						  if (g_ucTxKeyCode & M_OUT_BIT2){
								    //
									// b2 == 1 - pulse Data1 line
									//
								    ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA1, WGND_OUT_DATA1);
								  }
								  else {
								    //
									// b2 == 0 - pulse Data0 line
									//
								    ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA0, WGND_OUT_DATA0);
								  }
								  //
		                          // Start counting duration of Data0/Data1 line bit pulse
		                          //
		                          ROM_TimerEnable(WGND_OUT_TIM_BASE, WGND_TIMER);
								  //
		                          // Count sent bits
		                          //
								  g_ucOutBitCounter++;
		  						  break;

		  case WGNDOUT_TxBIT1	: //
		                          // Transmit bit b1
								  //
								  if (g_ucTxKeyCode & M_OUT_BIT1){
								    //
									// b1 == 1 - pulse Data1 line
									//
								    ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA1, WGND_OUT_DATA1);
								  }
								  else {
								    //
									// b1 == 0 - pulse Data0 line
									//
								    ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA0, WGND_OUT_DATA0);
								  }
								  //
		                          // Start counting duration of Data0/Data1 line bit pulse
		                          //
		                          ROM_TimerEnable(WGND_OUT_TIM_BASE, WGND_TIMER);
								  //
		                          // Count sent bits
		                          //
								  g_ucOutBitCounter++;
		  						  break;

		  case WGNDOUT_TxBIT0	: //
		                          // Transmit bit b0
								  //
								  if (g_ucTxKeyCode & M_OUT_BIT0){
								  	//
									// b0 == 1 - pulse Data1 line
								    //
									ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA1, WGND_OUT_DATA1);
								  }
								  else {
								    //
									// b0 == 0 - pulse Data0 line
								    //
									ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA0, WGND_OUT_DATA0);
								  }
								  //
		                          // Start counting duration of Data0/Data1 line bit pulse
		                          //
		                          ROM_TimerEnable(WGND_OUT_TIM_BASE, WGND_TIMER);
								  //
		                          // Count sent bits
		                          //
								  g_ucOutBitCounter++;
		  						  break;

		  case WGNDOUT_TxODD	: //
		                          // Transmit odd parity bit
								  //
								  if (((g_ucTxKeyCode & M_OUT_BIT1)>>1) == (g_ucTxKeyCode & M_OUT_BIT0)){
									  //
									  // b1 == b0 therefore odd parity bit = 1 - pulse Data1 line
									  //
									  ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA1, WGND_OUT_DATA1);
								  }
								  else{
								    //
									// b1 != b0 therefore odd parity bit = 0 - pulse Data0 line
									//
									ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA0, WGND_OUT_DATA0);
								  }
								  //
		                          // Start counting duration of Data0/Data1 line bit pulse
		                          //
		                          ROM_TimerEnable(WGND_OUT_TIM_BASE, WGND_TIMER);
								  //
		                          // Count sent bits
		                          //
								  g_ucOutBitCounter++;
		  						  break;

		  default				: //
		                          // End of frame mark
								  //
		                          if (g_ucOutBitCounter >= WGNDOUT_TxEND){
								    //
		                            // End of transmission - clear the transmission counter
		                            //
								    g_ucOutBitCounter = 0;
								  }
								  else{
								    //
		                            // Transmission still in progress - count delay between bit frames
		                            //
								    g_ucOutBitCounter++;
								  } 
		                          break;
		}

        //
        // Wait for the required amount of time.
        //
        xTaskDelayUntil(&ulLastTime, g_xWGNDoutBitDelay);
    }
}

//*****************************************************************************
//
// Initializes Wiegand outputs service task
//
//*****************************************************************************
unsigned long
WGNDoutTaskInit(void)
{
    //
    // Initialize the GPIO used as Wiegand outputs (2mA drive)
    //
    ROM_SysCtlPeripheralEnable(WGND_OUT_P_PERIPH);
    ROM_GPIOPinTypeGPIOOutput(WGND_OUT_P_BASE, WGND_OUT_DATA0 | WGND_OUT_DATA1);

    //
	// Set Wiegand outputs to idle state
    //
	ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA0 | WGND_OUT_DATA0, 0);

    //
	// Initialize the timer used to count duration of the data bit pulse
	//
    ROM_SysCtlPeripheralEnable(WGND_OUT_TIM_PERIPH);

    //
    // Configure the timer to trigger the interrupt after 50us
    //
    ROM_TimerConfigure(WGND_OUT_TIM_BASE, (TIMER_CFG_16_BIT_PAIR |
                                           TIMER_CFG_A_ONE_SHOT | TIMER_CFG_B_ONE_SHOT));
    ROM_TimerLoadSet(WGND_OUT_TIM_BASE, WGND_TIMER, ROM_SysCtlClockGet() / (1000000 / WGND_OUT_PULSE));

    //
    // Set the interrupts for the timer timeout
    //
    ROM_IntEnable(WGND_TIMER_INT);
    ROM_TimerIntEnable(WGND_OUT_TIM_BASE, WGND_TIMER_TMEOUT);

    //
    // Set the timer interrupt priority to 4 (i.e. no interrupt delay by kernel)
    //
    ROM_IntPrioritySet(WGND_TIMER_INT, 0x80);

    //
    // Create a queue for sending messages to the WGNDout task
    //
    if(xQueueCreate(g_cWGNDoutQueueBuffer, WGNDOUT_QUEUE_MEM_SIZE,
                    WGNDOUT_QUEUE_LENGTH, WGNDOUT_ITEM_SIZE, &g_pWGNDoutQueue) != pdPASS)
    {
        return(1);
    }

    //
    // Create the Wiegand outputs service task
    //
    if(xTaskCreate(WGNDoutTask, (signed portCHAR *)"WGNDout",
                   (signed portCHAR *)g_pulWGNDoutTaskStack,
                   sizeof(g_pulWGNDoutTaskStack), NULL, PRIORITY_WGNDout_TASK,
                   NULL) != pdPASS)
    {
        return(1);
    }

    //
    // Success.
    //
    return(0);
}

//*****************************************************************************
//
// The interrupt handler for the timer interrupt
//
// The handler simply resets Wiegand outputs to their idle state. In order to
// save CPU time, data is written directly to CPU registers rather than via
// Stellaris Peripheral Driver Library functions 
//
//*****************************************************************************
void
Timer0AIntHandler(void)
{
    //
    // Clear the timer interrupt
    //
    //ROM_TimerIntClear(WGND_OUT_TIM_BASE, WGND_TIMER_TMEOUT);
    HWREG(WGND_OUT_TIM_BASE + TIMER_O_ICR) = WGND_TIMER_TMEOUT;

    //
    // Reset Wiegand outputs to idle state
    //
	//ROM_GPIOPinWrite(WGND_OUT_P_BASE, WGND_OUT_DATA0 | WGND_OUT_DATA1, 0);
	HWREG(WGND_OUT_P_BASE + (GPIO_O_DATA + ((WGND_OUT_DATA0 | WGND_OUT_DATA1) << 2))) = 0;
}
