//*****************************************************************************
//
// display_task.c - The keypad pattern display service task
//
//*****************************************************************************

#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "driverlib/gpio.h"
#include "driverlib/rom.h"
#include "driverlib/sysctl.h"
#include "SafeRTOS/SafeRTOS_API.h"
#include "display_task.h"
#include "priorities.h"

//*****************************************************************************
//
// The GPIO pins to which the 7-segment displays are connected
//
//*****************************************************************************
// Display segments
#define DISP_SEG_P_PERIPH      SYSCTL_PERIPH_GPIOJ
#define DISP_SEG_P_BASE        GPIO_PORTJ_BASE
#define DISP_SEG_A             GPIO_PIN_0
#define DISP_SEG_B             GPIO_PIN_1
#define DISP_SEG_C             GPIO_PIN_2
#define DISP_SEG_D             GPIO_PIN_3
#define DISP_SEG_E             GPIO_PIN_4
#define DISP_SEG_F             GPIO_PIN_5
#define DISP_SEG_G             GPIO_PIN_6

// Display anodes
#define DISP_ANOD1_6_P_PERIPH  SYSCTL_PERIPH_GPIOH
#define DISP_ANOD1_6_P_BASE    GPIO_PORTH_BASE
#define DISP_ANODE1            GPIO_PIN_0
#define DISP_ANODE2            GPIO_PIN_1
#define DISP_ANODE3            GPIO_PIN_2
#define DISP_ANODE4            GPIO_PIN_5
#define DISP_ANODE5            GPIO_PIN_6
#define DISP_ANODE6            GPIO_PIN_7

#define DISP_ANOD7_12_P_PERIPH SYSCTL_PERIPH_GPIOD
#define DISP_ANOD7_12_P_BASE   GPIO_PORTD_BASE
#define DISP_ANODE7            GPIO_PIN_1
#define DISP_ANODE8            GPIO_PIN_2
#define DISP_ANODE9            GPIO_PIN_3
#define DISP_ANODE10           GPIO_PIN_4
#define DISP_ANODE11           GPIO_PIN_5
#define DISP_ANODE12           GPIO_PIN_6

//*****************************************************************************
//
// The amount of time to delay between successive digit refresh
//
//*****************************************************************************
const portTickType g_xDisplayRefreshDelay = 2;

//*****************************************************************************
//
// The display fonts map
//
//*****************************************************************************

// 
// Macro definition of 8-bit binary constant
//

// -> macro for long unsigned hex constant
#define HEX__(n) 0x##n##LU

// -> 8-bit conversion function
#define B8__(x) ((x&0x0000000FLU)?1:0) \
+((x&0x000000F0LU)?2:0) \
+((x&0x00000F00LU)?4:0) \
+((x&0x0000F000LU)?8:0) \
+((x&0x000F0000LU)?16:0) \
+((x&0x00F00000LU)?32:0) \
+((x&0x0F000000LU)?64:0) \
+((x&0xF0000000LU)?128:0)

// -> macro for 8-bit binary constant
#define B8(d) ((unsigned char)B8__(HEX__(d)))

//
// The display fonts map
//
const unsigned char g_ucDispFontROM[] =
{// segments: -gfedcba (binary value)
           B8(00111111),	// 0
		   B8(00000110),	// 1
		   B8(01011011),	// 2
		   B8(01001111),	// 3
		   B8(01100110),	// 4
		   B8(01101101),	// 5
		   B8(01111101),	// 6
		   B8(00000111),	// 7
		   B8(01111111),	// 8
		   B8(01101111),	// 9
		   B8(01011000),	// c - key [*]
		   B8(01011100),	// o - key [#]
		   B8(00000000)		// blank
};

//*****************************************************************************
//
// The stack for the display task
//
//*****************************************************************************
static unsigned long g_pulDisplayTaskStack[128];

//*****************************************************************************
//
// The keypad display buffer
//
//*****************************************************************************
static unsigned char g_ucDisplayBuffer[DISPLAY_SIZE];

//*****************************************************************************
//
// The display digit Index
//
//*****************************************************************************
static unsigned char g_ucDisplayDigitIndex = 0;

//*****************************************************************************
//
// The item size, queue size, and memory size for the display input queue
//
//*****************************************************************************
#define DISPLAY_ITEM_SIZE		sizeof(g_ucDisplayBuffer)
#define DISPLAY_QUEUE_LENGTH	1
#define DISPLAY_QUEUE_MEM_SIZE	((DISPLAY_ITEM_SIZE * DISPLAY_QUEUE_LENGTH) + \
                                 portQUEUE_OVERHEAD_BYTES)

//*****************************************************************************
//
// A buffer to contain the contents of the display input queue
//
//*****************************************************************************
signed portCHAR g_cDisplayQueueBuffer[DISPLAY_QUEUE_MEM_SIZE];

//*****************************************************************************
//
// The queue that holds messages sent to the display task
//
//*****************************************************************************
static xQueueHandle g_pDisplayQueue;

//*****************************************************************************
//
// Maximum delay to display the new key pattern
//
//*****************************************************************************
#define DISPLAY_MAX_DELAY	25

//*****************************************************************************
//
// Sends a request to the display task to display a new key pattern
//
//*****************************************************************************
void
DisplayKeyPattern(unsigned char *pucKeyPattern)
{
    //
    // Send the display key pattern request to the display task
    //
    xQueueSend(g_pDisplayQueue, pucKeyPattern, DISPLAY_MAX_DELAY);
}

//*****************************************************************************
//
// This task services the keypad pattern 7-segment displays. Each digit is
// displayed for 2ms, therefore it takes 24ms (12*2ms) to refresh all displays.
//
//*****************************************************************************
static void
DisplayTask(void *pvParameters)
{
    portTickType ulLastTime;

    //
    // Get the current tick count.
    //
    ulLastTime = xTaskGetTickCount();

    //
    // Loop forever.
    //
    while(1)
    {
		//
		// Check the display input queue
		// (The queue is checked at the beginning of the display refresh cycle)
		//
		if (g_ucDisplayDigitIndex == 0){
		  xQueueReceive(g_pDisplayQueue, g_ucDisplayBuffer, 0);  
		}

		//
		// Select the 7-segment display
		//
		ROM_GPIOPinWrite(DISP_ANOD1_6_P_BASE,
		                 DISP_ANODE1 |DISP_ANODE2 | DISP_ANODE3 | DISP_ANODE4 | DISP_ANODE5 |DISP_ANODE6,
						 0x00);				// turn off anodes No.1..6
		ROM_GPIOPinWrite(DISP_ANOD7_12_P_BASE,
		                 DISP_ANODE7 | DISP_ANODE8 | DISP_ANODE9 |DISP_ANODE10 | DISP_ANODE11 | DISP_ANODE12,
						 0x00);				// turn off anodes No.7..12

		switch (g_ucDisplayDigitIndex){
		  case 0  :   // turn on anode No.1
		              ROM_GPIOPinWrite(DISP_ANOD1_6_P_BASE, DISP_ANODE1, DISP_ANODE1);
		              break;

		  case 1  :   // turn on anode No.2
		              ROM_GPIOPinWrite(DISP_ANOD1_6_P_BASE, DISP_ANODE2, DISP_ANODE2);
		              break;

		  case 2  :   // turn on anode No.3
		              ROM_GPIOPinWrite(DISP_ANOD1_6_P_BASE, DISP_ANODE3, DISP_ANODE3);
		              break;
		  
		  case 3  :   // turn on anode No.4
		              ROM_GPIOPinWrite(DISP_ANOD1_6_P_BASE, DISP_ANODE4, DISP_ANODE4);
		              break;

		  case 4  :   // turn on anode No.5
		              ROM_GPIOPinWrite(DISP_ANOD1_6_P_BASE, DISP_ANODE5, DISP_ANODE5);
		              break;

		  case 5  :   // turn on anode No.6
		              ROM_GPIOPinWrite(DISP_ANOD1_6_P_BASE, DISP_ANODE6, DISP_ANODE6);
		              break;

		  case 6  :   // turn on anode No.7
		              ROM_GPIOPinWrite(DISP_ANOD7_12_P_BASE, DISP_ANODE7, DISP_ANODE7);
		              break;

		  case 7  :   // turn on anode No.8
		              ROM_GPIOPinWrite(DISP_ANOD7_12_P_BASE, DISP_ANODE8, DISP_ANODE8);
		              break;

		  case 8  :   // turn on anode No.9 
		              ROM_GPIOPinWrite(DISP_ANOD7_12_P_BASE, DISP_ANODE9, DISP_ANODE9);
		              break;

		  case 9  :   // turn on anode No.10
		              ROM_GPIOPinWrite(DISP_ANOD7_12_P_BASE, DISP_ANODE10, DISP_ANODE10);
		              break;

		  case 10 :   // turn on anode No.11
		              ROM_GPIOPinWrite(DISP_ANOD7_12_P_BASE, DISP_ANODE11, DISP_ANODE11);
		              break;

		  case 11 :   // turn on anode No.12
		              ROM_GPIOPinWrite(DISP_ANOD7_12_P_BASE, DISP_ANODE12, DISP_ANODE12);
		              break;

		  default :   break;
		}

		//
		// Drive the display segments
		//
		ROM_GPIOPinWrite(DISP_SEG_P_BASE,
		                 DISP_SEG_A |DISP_SEG_B | DISP_SEG_C | DISP_SEG_D | DISP_SEG_E |DISP_SEG_F | DISP_SEG_G,
						 g_ucDispFontROM[g_ucDisplayBuffer[g_ucDisplayDigitIndex]]);

		//
		// Move the display refresh index to the next position
		//
		g_ucDisplayDigitIndex++;
		if (g_ucDisplayDigitIndex >= DISPLAY_SIZE) g_ucDisplayDigitIndex=0;

        //
        // Wait for the required amount of time.
        //
        xTaskDelayUntil(&ulLastTime, g_xDisplayRefreshDelay);
    }
}

//*****************************************************************************
//
// Initializes the keypad display service task
//
//*****************************************************************************
unsigned long
DisplayTaskInit(void)
{
    //
    // Initialize the GPIO used to drive the display segments (2mA drive)
    //
    ROM_SysCtlPeripheralEnable(DISP_SEG_P_PERIPH);
    ROM_GPIOPinTypeGPIOOutput(DISP_SEG_P_BASE,
	  DISP_SEG_A |DISP_SEG_B | DISP_SEG_C | DISP_SEG_D | DISP_SEG_E |DISP_SEG_F | DISP_SEG_G);

    //
    // Initialize the GPIO used to drive the display anodes (2mA drive)
    //
    ROM_SysCtlPeripheralEnable(DISP_ANOD1_6_P_PERIPH);
    ROM_GPIOPinTypeGPIOOutput(DISP_ANOD1_6_P_BASE,
	  DISP_ANODE1 |DISP_ANODE2 | DISP_ANODE3 | DISP_ANODE4 | DISP_ANODE5 |DISP_ANODE6);

    ROM_SysCtlPeripheralEnable(DISP_ANOD7_12_P_PERIPH);
    ROM_GPIOPinTypeGPIOOutput(DISP_ANOD7_12_P_BASE,
	  DISP_ANODE7 | DISP_ANODE8 | DISP_ANODE9 |DISP_ANODE10 | DISP_ANODE11 | DISP_ANODE12);

    //
    // Create a queue for sending messages to the display task
    //
    if(xQueueCreate(g_cDisplayQueueBuffer, DISPLAY_QUEUE_MEM_SIZE,
                    DISPLAY_QUEUE_LENGTH, DISPLAY_ITEM_SIZE, &g_pDisplayQueue) != pdPASS)
    {
        return(1);
    }

    //
    // Create the keypad display service task
    //
    if(xTaskCreate(DisplayTask, (signed portCHAR *)"Display",
                   (signed portCHAR *)g_pulDisplayTaskStack,
                   sizeof(g_pulDisplayTaskStack), NULL, PRIORITY_DISPLAY_TASK,
                   NULL) != pdPASS)
    {
        return(1);
    }

    //
    // Success.
    //
    return(0);
}
