//*****************************************************************************
//
// reader_task.c - The reader task
//
//*****************************************************************************

#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "driverlib/gpio.h"
#include "driverlib/rom.h"
#include "driverlib/sysctl.h"
#include "SafeRTOS/SafeRTOS_API.h"
#include "reader_task.h"
#include "display_task.h"
#include "touch_task.h"
#include "WGNDin_task.h"
#include "WGNDout_task.h"
#include "priorities.h"
#include "uart_task.h"
#include "utils/ustdlib.h"

//*****************************************************************************
//
// The amount of time to service the reader task
//
//*****************************************************************************
const portTickType g_xReaderServiceTime = 1;

//*****************************************************************************
//
// The stack for the reader task
//
//*****************************************************************************
static unsigned long g_pulReaderTaskStack[128];

//*****************************************************************************
//
// The keypad patterns: display off, display init and display workspace
//
//*****************************************************************************
const unsigned char g_ucKeyPatternOff[DISPLAY_SIZE]=
  {DISP_BLNK, DISP_BLNK, DISP_BLNK,
   DISP_BLNK, DISP_BLNK, DISP_BLNK,
   DISP_BLNK, DISP_BLNK, DISP_BLNK,
   DISP_BLNK, DISP_BLNK, DISP_BLNK};

const unsigned char g_ucKeyPatternInit[DISPLAY_SIZE] =
  {DISP_1, DISP_2, DISP_3,
   DISP_4, DISP_5, DISP_6,
   DISP_7, DISP_8, DISP_9,
   DISP_c, DISP_0, DISP_o};

static unsigned char g_ucKeyPattern[DISPLAY_SIZE] =
  {DISP_1, DISP_2, DISP_3,
   DISP_4, DISP_5, DISP_6,
   DISP_7, DISP_8, DISP_9,
   DISP_c, DISP_0, DISP_o};

//*****************************************************************************
//
// The pressed key coordinates
//
//*****************************************************************************
TP_position g_KeyXY;

//*****************************************************************************
//
// The pressed key code
//
//*****************************************************************************
static unsigned char g_ucKeyCode;

//*****************************************************************************
//
// The keypad display turn off timer
//
//*****************************************************************************
#define KEYPAD_OFF_TIME		10000
static unsigned short g_usKeypadOffTimer = KEYPAD_OFF_TIME;

//*****************************************************************************
//
// The random number
//
//*****************************************************************************
static unsigned long g_ulRandomNumber = 0x01;

//*****************************************************************************
//
// This function computes pseudo-random number. The algorithm of random number
// generation is based on LFSF (i.e. Line Feedback Shift Register) technique.
// The function uses 31-bit register with feedback taps at bits 2 and 30 which
// results in pseudo-random sequence of length 2 147 483 647.  
//
//*****************************************************************************
static unsigned long
GetRandomNumber(void)
{
    unsigned long RndTemp;						// scratchpad register

	//
	// Load previous shift register state
	//
	RndTemp = g_ulRandomNumber << 1;

	//
	// Compute feedback bit
	//
    if (((RndTemp & 0x80000080) == 0x80000000)	// b30=1 and b2=0
	 || ((RndTemp & 0x80000080) == 0x00000080)){// or b30=0 and b2=1
	  //
	  // b30 != b2 - feedback bit equals 1
	  //	  
	  RndTemp |= 0x00000001;
	}

	//
	// Clear unused bit	b31
	//
	RndTemp &= 0x7FFFFFFF;

	//
	// Store current shift register state
	//
	g_ulRandomNumber = RndTemp;
	return (RndTemp);
}

//*****************************************************************************
//
// This function computes random permutation of the keypad pattern. Function
// keys [*] and [#] are always left in place for ease of access.
//    
// The implemented algorithm of random permutation comes from R.Grig's blog
// "R.Grig's Thoughts", chapter "Theory and Practice", entry "Random Permutation".
// The whole idea of computing random permutation of the set of N digits from 0 to
// N-1 is very simple. Lets assume that array p[i] contains permutation of numbers
// from 0 to i-1. Now we choose random position in this array and write there number
// i, moving the old value from this position to position p[i]. As the result we
// receive random permutation of the set of i numbers from 0 to i.
//
//*****************************************************************************
static void
GenerateKeypadPattern(unsigned char *pKeyPatternBuffer)
{
    unsigned char i, j;				// loop counters

	//
	// Set the first number of keypad pattern permutation
	//
    *pKeyPatternBuffer = g_ucKeyPatternInit[0];

	//
	// Compute keypad pattern permutation
	//
	for (i = 0; i < DISPLAY_SIZE-2; ++i){
	  //
	  // Get the random position in the array
	  //
	  j = GetRandomNumber() % (i+1);
	  //
	  // Move numbers in the array (skip the [*] key)
	  //
	  *(pKeyPatternBuffer + i) = *(pKeyPatternBuffer + j);
	  *(pKeyPatternBuffer + j) = g_ucKeyPatternInit[((i != DISPLAY_SIZE-3) ? i : i+1)];
	}
	//
	// Make room for the [*] key
	//
	*(pKeyPatternBuffer + DISPLAY_SIZE-2) = *(pKeyPatternBuffer + DISPLAY_SIZE-3);
	//
	// Write the [*] key (no change in its position)
	//
	*(pKeyPatternBuffer + DISPLAY_SIZE-3) = g_ucKeyPatternInit[DISPLAY_SIZE-3];
	//
	// Write the [#] key (no change in its position)
	//
	*(pKeyPatternBuffer + DISPLAY_SIZE-1) = g_ucKeyPatternInit[DISPLAY_SIZE-1];
}

//*****************************************************************************
//
// This task integrates independent tasks servicing the reader peripherals into
// one coherent program. Because integrated tasks provide input data periodically,
// with time step of 1ms or more, there is no use of servicing the reader task
// more often than every 1ms.
//
//*****************************************************************************
static void
ReaderTask(void *pvParameters)
{
    portTickType ulLastTime;
	char pcXYString[16];			// string conversion buffer
	char temp;						// scratchpad register

	//
    // Get the current tick count
    //
    ulLastTime = xTaskGetTickCount();

	//
	// Display start keypad pattern
	//
    DisplayKeyPattern(g_ucKeyPattern);

    //
    // Loop forever.
    //
    while(1)
    {
		//
		// Check the touch panel output queue
		//
		if (CheckTouchPanel(&g_KeyXY) != 0){

		  //
		  // The touch panel has been pressed - check the keypad pattern display
		  //
          if (g_usKeypadOffTimer == 0){

		    //
			// The keypad pattern display is off - generate new keypad pattern and turn the display on
			//
			GenerateKeypadPattern(g_ucKeyPattern);
		    DisplayKeyPattern(g_ucKeyPattern);
		  }

		  else{

		    //
			// The keypad pattern display is on - process touch coordinates
			//
			if ((g_KeyXY.Xcol == 0) || (g_KeyXY.Yrow == 0)){

 			  //
			  // The pen is up - light the pressed digit
			  //
			  DisplayKeyPattern(g_ucKeyPattern);
			  
			  //
			  // Check if the user requested for a new keypad pattern
			  //
			  if (g_ucKeyCode == DISP_c){

			  //
              // Compute new keypad pattern
			  //
			  GenerateKeypadPattern(g_ucKeyPattern);
		      DisplayKeyPattern(g_ucKeyPattern);

			  //
			  // Clear code of the pressed key
			  //
			  g_ucKeyCode = DISP_BLNK;
			  }
			}

			else{

			  //
			  // The pen is down - convert touch coordinates to the key pattern
			  // table index and read code of the pressed key
			  //
			  temp = (g_KeyXY.Yrow - 1)*3 + (g_KeyXY.Xcol - 1);
			  g_ucKeyCode = g_ucKeyPattern[temp];
			  
			  //
			  // Dim the pressed digit
			  // (temp - key pattern table index of the pressed key)
			  //
			  g_ucKeyPattern[temp] = DISP_BLNK;
			  DisplayKeyPattern(g_ucKeyPattern);
			  g_ucKeyPattern[temp] = g_ucKeyCode;

			  //
			  // Confirm the touch panel press
			  //
			  MakeBeep(BEEP_TIME);

			  //
			  // Send the key code to the ACC system
			  //
			  WGNDOutputKeyCode(g_ucKeyCode);

			  //
              // Send the key code message to the UART queue
              //
			  UARTOutputString("Key code: 0x");
			  //
              // Convert the X,Y position into a string
              //
			  usnprintf(pcXYString, sizeof(pcXYString), "%02X \n", g_ucKeyCode);
              //
              // Display this string
              //
              UARTOutputString(pcXYString);
			}
		  }

		  //
		  // Restart keypad pattern display off timer
		  //
		  g_usKeypadOffTimer = KEYPAD_OFF_TIME;
		}

		//
		// Service keypad pattern display off timer
		//
		if (g_usKeypadOffTimer != 0){
		  //
		  // The keypad pattern display is on - count time to turn it off
		  //
		  g_usKeypadOffTimer--;
		  if (g_usKeypadOffTimer == 0){
		    //
			// Turn the keypad pattern display off
			//
			DisplayKeyPattern((unsigned char *)g_ucKeyPatternOff); 
		  }
		} 

        //
        // Wait for the required amount of time.
        //
        xTaskDelayUntil(&ulLastTime, g_xReaderServiceTime);
    }
}

//*****************************************************************************
//
// Initializes the reader task
//
//*****************************************************************************
unsigned long
ReaderTaskInit(void)
{
    //
    // Create the reader task
    //
    if(xTaskCreate(ReaderTask, (signed portCHAR *)"Reader",
                   (signed portCHAR *)g_pulReaderTaskStack,
                   sizeof(g_pulReaderTaskStack), NULL, PRIORITY_READER_TASK,
                   NULL) != pdPASS)
    {
        return(1);
    }

    //
    // Success.
    //
    return(0);
}
