//*****************************************************************************
//
// touch_task.c - The touch panel service task
//
// Code of this module is based on the touch screen driver for the DK-LM3S9B96
// board, which is part of the DK-LM3S9B96 Firmware Package, revision 5570.
// The main modifications of the original 'touch.c' driver include:
//  - Stellaris Peripheral Driver Library functions are called from ROM,
//  - ADC conversions are triggered by software (i.e. the touch panel service
//    task) rather than by Timer0,
//  - ADC conversion results are analyzed by the touch panel service task, not
//    in ADC interrupt service routine,
//  - calibration parameters are adjusted to the mobile phone LG KU990 touch
//    panel that operates in portrait mode
//
//*****************************************************************************

#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "inc/hw_adc.h"
#include "inc/hw_gpio.h"
#include "driverlib/gpio.h"
#include "driverlib/rom.h"
#include "driverlib/sysctl.h"
#include "driverlib/adc.h"
#include "SafeRTOS/SafeRTOS_API.h"
#include "touch_task.h"
#include "priorities.h"
#include "uart_task.h"
#include "utils/ustdlib.h"

//*****************************************************************************
//
// The GPIO pins to which the touch panel is connected
//
//*****************************************************************************
#define TS_P_PERIPH             SYSCTL_PERIPH_GPIOE
#define TS_P_BASE               GPIO_PORTE_BASE
#define TS_N_PERIPH             SYSCTL_PERIPH_GPIOE
#define TS_N_BASE               GPIO_PORTE_BASE
#define TS_XP_PIN               GPIO_PIN_6
#define TS_YP_PIN               GPIO_PIN_7
#define TS_XN_PIN               GPIO_PIN_2
#define TS_YN_PIN               GPIO_PIN_3

//*****************************************************************************
//
// The ADC channels connected to each of the touch panel contacts
//
//*****************************************************************************
#define ADC_CTL_CH_XP           ADC_CTL_CH1
#define ADC_CTL_CH_YP           ADC_CTL_CH0

//*****************************************************************************
//
// The amount of time to delay between successive touch panel ADC samples
//
//*****************************************************************************
const portTickType g_xTouchPanelSampleDelay = 1;

//*****************************************************************************
//
// The coefficients used to convert from the ADC touch screen readings to the
// screen pixel positions
//
//*****************************************************************************
//*****************************************************************************
//
// Touch panel calibration parameters
// (The touch panel KU990 operates in portrait orientation:
//				    col1 col2 col3
//              row1
//              row2
//              row3
//              row4
//  There are two column threshold levels and three row threshold levels
//
//*****************************************************************************
#define COL_1_2		0x210
#define COL_2_3		0x2CF

#define ROW_1_2		0x190
#define ROW_2_3		0x258
#define ROW_3_4		0x308

//*****************************************************************************
//
// The lowest ADC reading assumed to represent a press on the screen.  Readings
// below this indicate no press is taking place.
//
//*****************************************************************************
const unsigned short g_sTouchMin = 150;

//*****************************************************************************
//
// Panel touch position
//
//*****************************************************************************
TP_position TPKeyPos;

//*****************************************************************************
//
// The stack for the touch panel task
//
//*****************************************************************************
static unsigned long g_pulTouchPanelTaskStack[128];

//*****************************************************************************
//
// The item size, queue size, and memory size for the touch panel output queue
//
//*****************************************************************************
#define TP_ITEM_SIZE			sizeof(TPKeyPos)
#define TP_QUEUE_LENGTH			1
#define TP_QUEUE_MEM_SIZE		((TP_ITEM_SIZE * TP_QUEUE_LENGTH) + \
                                 portQUEUE_OVERHEAD_BYTES)

//*****************************************************************************
//
// A buffer to contain the contents of the touch panel output queue
//
//*****************************************************************************
signed portCHAR g_cTPQueueBuffer[TP_QUEUE_MEM_SIZE];

//*****************************************************************************
//
// The queue that holds messages sent out by the touch panel
//
//*****************************************************************************
static xQueueHandle g_pTPQueue;

//*****************************************************************************
//
// The current state of the touch screen driver's state machine.  This is used
// to cycle the touch screen interface through the powering sequence required
// to read the two axes of the surface.
//
//*****************************************************************************
static unsigned long g_ulTSState;
#define TS_STATE_INIT           0
#define TS_STATE_READ_X         1
#define TS_STATE_READ_Y         2
#define TS_STATE_SKIP_X         3
#define TS_STATE_SKIP_Y         4

//*****************************************************************************
//
// The most recent raw ADC reading for the X position on the screen.  This
// value is not affected by the selected screen orientation.
//
//*****************************************************************************
volatile short g_sTouchX;

//*****************************************************************************
//
// The most recent raw ADC reading for the Y position on the screen.  This
// value is not affected by the selected screen orientation.
//
//*****************************************************************************
volatile short g_sTouchY;

//*****************************************************************************
//
// A pointer to the function to receive messages from the touch screen driver
// when events occur on the touch screen (debounced presses, movement while
// pressed, and debounced releases).
//
//*****************************************************************************
//static long (*g_pfnTSHandler)(unsigned long ulMessage, long lX, long lY);

//*****************************************************************************
//
// The current state of the touch screen debouncer.  When zero, the pen is up.
// When three, the pen is down.  When one or two, the pen is transitioning from
// one state to the other.
//
//*****************************************************************************
static unsigned char g_cState = 0;

//*****************************************************************************
//
// The queue of debounced pen positions.  This is used to slightly delay the
// returned pen positions, so that the pen positions that occur while the pen
// is being raised are not send to the application.
//
//*****************************************************************************
static short g_psSamples[8];

//*****************************************************************************
//
// The count of pen positions in g_psSamples.  When negative, the buffer is
// being pre-filled as a result of a detected pen down event.
//
//*****************************************************************************
static signed char g_cIndex = 0;

//*****************************************************************************
//
// Checks the touch panel output queue
//
//*****************************************************************************
char
CheckTouchPanel(TP_position *pKeyPos)
{
    //
    // Read the message from the queue
    //
    if(xQueueReceive(g_pTPQueue, pKeyPos, 0) != pdPASS)
    {
        return(0);
    }

    //
    // Success.
    //
    return(1);
}

//*****************************************************************************
//
// Debounces presses of the touch screen.
//
// This function is called when a new X/Y sample pair has been captured in
// order to perform debouncing of the touch screen.
//
// \return None.
//
//*****************************************************************************
static void
TouchScreenDebouncer(void)
{
    char pcXYString[16];			// string conversion buffer

    //
    // Convert the ADC readings into key values on the screen
    //

    // - X position
	if (g_sTouchX < COL_1_2) TPKeyPos.Xcol = 1;			// column no.1
	  else if (g_sTouchX < COL_2_3) TPKeyPos.Xcol = 2;	// column no.2
	  else TPKeyPos.Xcol = 3;							// column no.3

    // - Y position
	if (g_sTouchY < ROW_1_2) TPKeyPos.Yrow = 1;			// row no.1
	  else if (g_sTouchY < ROW_2_3) TPKeyPos.Yrow = 2;	// row no.2
	  else if (g_sTouchY < ROW_3_4) TPKeyPos.Yrow = 3;	// row no.3
	  else TPKeyPos.Yrow = 4;							// row no.4

    //
    // See if the touch screen is being touched
    //
    if((g_sTouchX < g_sTouchMin) || (g_sTouchY < g_sTouchMin))
    {
        //
        // See if the pen is not up right now
        //
        if(g_cState != 0x00)
        {
            //
            // Decrement the state count
            //
            g_cState--;

            //
            // See if the pen has been detected as up three times in a row
            //
            if(g_cState == 0x80)
            {
                //
                // Indicate that the pen is up
                //
                g_cState = 0x00;

				//
                // Send the pen up message to the touch screen output queue
                //
				TPKeyPos.Xcol = 0;
				TPKeyPos.Yrow = 0;
				xQueueSend(g_pTPQueue, &TPKeyPos, 0);

				//
                // Send the pen up message to the UART queue
                //
				UARTOutputString("Pen is up \n");
            }
        }
    }
    else
    {
        //
        // See if the pen is not down right now
        //
        if(g_cState != 0x83)
        {
            //
            // Increment the state count
            //
            g_cState++;

            //
            // See if the pen has been detected as down three times in a row
            //
            if(g_cState == 0x03)
            {
                //
                // Indicate that the pen is up
                //
                g_cState = 0x83;

                //
                // Set the index to -8, so that the next 3 samples are stored
                // into the sample buffer before sending anything back to the
                // touch screen event handler.
                //
                g_cIndex = -8;

                //
                // Store this sample into the sample buffer
                //
                g_psSamples[0] = TPKeyPos.Xcol;
                g_psSamples[1] = TPKeyPos.Yrow;
            }
        }
        else
        {
            //
            // See if the sample buffer pre-fill has completed
            //
            if(g_cIndex == -2)
            {
                //
                // Send the pen down message to the touch screen output queue
                //
				//
				xQueueSend(g_pTPQueue, &TPKeyPos, 0);

				//
                // Send the pen up message to the UART queue
                //
				UARTOutputString("Pen is down at XY:");
				//
                // Convert the X,Y position into a string
                //
				usnprintf(pcXYString, sizeof(pcXYString), "%2X %2X \n", TPKeyPos.Xcol, TPKeyPos.Yrow);
                //
                // Display this string
                //
                UARTOutputString(pcXYString);

                //
                // Store this sample into the sample buffer
                //
                g_psSamples[0] = TPKeyPos.Xcol;
                g_psSamples[1] = TPKeyPos.Yrow;

                //
                // Set the index to the next sample to send
                //
                g_cIndex = 2;
            }

            //
            // Otherwise, see if the sample buffer pre-fill is in progress
            //
            else if(g_cIndex < 0)
            {
                //
                // Store this sample into the sample buffer
                //
                g_psSamples[g_cIndex + 10] = TPKeyPos.Xcol;
                g_psSamples[g_cIndex + 11] = TPKeyPos.Yrow;

                //
                // Increment the index
                //
                g_cIndex += 2;
            }

            //
            // Otherwise, the sample buffer is full
            //
            else
            {
                //
                // Send the pen move message to the touch screen output queue
				// (not supported in this application)
                //
                //UARTOutputString("Pen is moved to XY:");
				//
                // Convert the X,Y position into a string
                //
				//usnprintf(pcXYString, sizeof(pcXYString), "%2x %2x \n", XYcol, XYrow);
                //
                // Display this string
                //
                //UARTOutputString(pcXYString);

                //
                // Store this sample into the sample buffer
                //
                g_psSamples[g_cIndex] = TPKeyPos.Xcol;
                g_psSamples[g_cIndex + 1] = TPKeyPos.Yrow;

                //
                // Increment the index
                //
                g_cIndex = (g_cIndex + 2) & 7;
            }
        }
    }
}

//*****************************************************************************
//
// This task services the touch panel. The touch screen state machine is advanced
// and the acquired ADC sample is processed.
//
//*****************************************************************************
static void
TouchPanelTask(void *pvParameters)
{
    portTickType ulLastTime;

    //
    // Get the current tick count
    //
    ulLastTime = xTaskGetTickCount();

    //
    // Loop forever
    //
    while(1)
    {
		//
    	// Determine what to do based on the current state of the state machine
    	//
    	switch(g_ulTSState)
    	{
          //
          // The new sample is an X axis sample that should be discarded
          //
          case TS_STATE_SKIP_X:
          {
            //
            // Set the analog mode select for the YP pin
            //
            HWREG(TS_P_BASE + GPIO_O_AMSEL) =
                HWREG(TS_P_BASE + GPIO_O_AMSEL) | TS_YP_PIN;

            //
            // Configure the Y axis touch layer pins as inputs
            //
            HWREG(TS_P_BASE + GPIO_O_DIR) =
                HWREG(TS_P_BASE + GPIO_O_DIR) & ~TS_YP_PIN;

            HWREG(TS_N_BASE + GPIO_O_DIR) =
                HWREG(TS_N_BASE + GPIO_O_DIR) & ~TS_YN_PIN;

            //
            // The next sample will be a valid X axis sample
            //
            g_ulTSState = TS_STATE_READ_X;

            //
		    // Start ADC conversion
		    //
            ROM_ADCProcessorTrigger(ADC0_BASE, 3);

            //
            // This state has been handled
            //
            break;
          }

          //
          // The new sample is an X axis sample that should be processed
          //
          case TS_STATE_READ_X:
          {
            //
            // Read the raw ADC sample
            //
            g_sTouchX = HWREG(ADC0_BASE + ADC_O_SSFIFO3);

            //
            // Clear the analog mode select for the YP pin
            //
            HWREG(TS_P_BASE + GPIO_O_AMSEL) =
                HWREG(TS_P_BASE + GPIO_O_AMSEL) & ~TS_YP_PIN;

            //
            // Configure the X and Y axis touch layers as outputs
            //
            HWREG(TS_P_BASE + GPIO_O_DIR) =
                HWREG(TS_P_BASE + GPIO_O_DIR) | TS_XP_PIN | TS_YP_PIN;

            HWREG(TS_N_BASE + GPIO_O_DIR) =
                HWREG(TS_N_BASE + GPIO_O_DIR) | TS_XN_PIN | TS_YN_PIN;

            //
            // Drive the positive side of the Y axis touch layer with VDD and
            // the negative side with GND.  Also, drive both sides of the X
            // axis layer with GND to discharge any residual voltage (so that
            // a no-touch condition can be properly detected).
            //
            HWREG(TS_N_BASE + GPIO_O_DATA + ((TS_XN_PIN | TS_YN_PIN) << 2)) = 0;
            
            HWREG(TS_P_BASE + GPIO_O_DATA + ((TS_XP_PIN | TS_YP_PIN) << 2)) =
                TS_YP_PIN;

            //
            // Configure the sample sequence to capture the X axis value
            //
            HWREG(ADC0_BASE + ADC_O_SSMUX3) = ADC_CTL_CH_XP;

            //
            // The next sample will be an invalid Y axis sample
            //
            g_ulTSState = TS_STATE_SKIP_Y;

            //
            // This state has been handled
            //
            break;
          }

          //
          // The new sample is a Y axis sample that should be discarded
          //
          case TS_STATE_SKIP_Y:
          {
            //
            // Set the analog mode select for the XP pin
            //
            HWREG(TS_P_BASE + GPIO_O_AMSEL) =
                HWREG(TS_P_BASE + GPIO_O_AMSEL) | TS_XP_PIN;

            //
            // Configure the X axis touch layer pins as inputs
            //
            HWREG(TS_P_BASE + GPIO_O_DIR) =
                HWREG(TS_P_BASE + GPIO_O_DIR) & ~TS_XP_PIN;

            HWREG(TS_N_BASE + GPIO_O_DIR) =
                HWREG(TS_N_BASE + GPIO_O_DIR) & ~TS_XN_PIN;
            
            //
            // The next sample will be a valid Y axis sample
            //
            g_ulTSState = TS_STATE_READ_Y;

            //
		    // Start ADC conversion
		    //
            ROM_ADCProcessorTrigger(ADC0_BASE, 3);

            //
            // This state has been handled
            //
            break;
          }

          //
          // The new sample is a Y axis sample that should be processed
          //
          case TS_STATE_READ_Y:
          {
            //
            // Read the raw ADC sample
            //
            g_sTouchY = HWREG(ADC0_BASE + ADC_O_SSFIFO3);

            //
            // The next configuration is the same as the initial configuration.
            // Therefore, fall through into the initialization state to avoid
            // duplicating the code.
            //
          }

          //
          // The state machine is in its initial state
          //
          case TS_STATE_INIT:
          {
            //
            // Clear the analog mode select for the XP pin
            //
            HWREG(TS_P_BASE + GPIO_O_AMSEL) =
                HWREG(TS_P_BASE + GPIO_O_AMSEL) & ~TS_XP_PIN;

            //
            // Configure the X and Y axis touch layers as outputs
            //
            HWREG(TS_P_BASE + GPIO_O_DIR) =
                HWREG(TS_P_BASE + GPIO_O_DIR) | TS_XP_PIN | TS_YP_PIN;

            HWREG(TS_N_BASE + GPIO_O_DIR) =
                HWREG(TS_N_BASE + GPIO_O_DIR) | TS_XN_PIN | TS_YN_PIN;

            //
            // Drive one side of the X axis touch layer with VDD and the other
            // with GND.  Also, drive both sides of the Y axis layer with GND
            // to discharge any residual voltage (so that a no-touch condition
            // can be properly detected).
            //
            HWREG(TS_P_BASE + GPIO_O_DATA + ((TS_XP_PIN | TS_YP_PIN) << 2)) =
                TS_XP_PIN;

            HWREG(TS_N_BASE + GPIO_O_DATA + ((TS_XN_PIN | TS_YN_PIN) << 2)) = 0;
            
            //
            // Configure the sample sequence to capture the Y axis value
            //
            HWREG(ADC0_BASE + ADC_O_SSMUX3) = ADC_CTL_CH_YP;

            //
            // If this is the valid Y sample state, then there is a new X/Y
            // sample pair. In that case, run the touch screen debouncer.
            //
            if(g_ulTSState == TS_STATE_READ_Y)
            {
                TouchScreenDebouncer();
            }

            //
            // The next sample will be an invalid X axis sample.
            //
            g_ulTSState = TS_STATE_SKIP_X;

            //
            // This state has been handled.
            //
            break;
          }
    	}

        //
        // Wait for the required amount of time.
        //
        xTaskDelayUntil(&ulLastTime, g_xTouchPanelSampleDelay);
    }
}

//*****************************************************************************
//
// Initializes the touch panel service task
//
//*****************************************************************************
unsigned long
TouchPanelTaskInit(void)
{
    //
    // Set the initial state of the touch screen driver's state machine
    //
    g_ulTSState = TS_STATE_INIT;

    //
    // Enable the peripherals used by the touch panel interface
    //
    ROM_SysCtlPeripheralEnable(SYSCTL_PERIPH_ADC0);
    ROM_SysCtlPeripheralEnable(TS_P_PERIPH);
    ROM_SysCtlPeripheralEnable(TS_N_PERIPH);

    //
    // Configure the ADC sample sequence used to read the touch panel reading
    //
    ROM_ADCHardwareOversampleConfigure(ADC0_BASE, 4);
    ROM_ADCSequenceConfigure(ADC0_BASE, 3, ADC_TRIGGER_PROCESSOR, 0);
    ROM_ADCSequenceStepConfigure(ADC0_BASE, 3, 0, ADC_CTL_CH_YP | ADC_CTL_IE | ADC_CTL_END);
    ROM_ADCSequenceEnable(ADC0_BASE, 3);

    //
    // Configure the GPIOs used to drive the touch panel layers
    //
    ROM_GPIOPinTypeGPIOOutput(TS_P_BASE, TS_XP_PIN | TS_YP_PIN);

    //
    // Set up GPIOs to drive the XN and YN signals
    //
    ROM_GPIOPinTypeGPIOOutput(TS_N_BASE, TS_XN_PIN | TS_YN_PIN);

    ROM_GPIOPinWrite(TS_P_BASE, TS_XP_PIN | TS_YP_PIN, 0x00);
    ROM_GPIOPinWrite(TS_N_BASE, TS_XN_PIN | TS_YN_PIN, 0x00);

    //
    // Create a queue for sending messages from the touch panel task
    //
    if(xQueueCreate(g_cTPQueueBuffer, TP_QUEUE_MEM_SIZE,
                    TP_QUEUE_LENGTH, TP_ITEM_SIZE, &g_pTPQueue) != pdPASS)
    {
        return(1);
    }

	//
    // Create the touch panel service task
    //
    if(xTaskCreate(TouchPanelTask, (signed portCHAR *)"TouchPanel",
                   (signed portCHAR *)g_pulTouchPanelTaskStack,
                   sizeof(g_pulTouchPanelTaskStack), NULL, PRIORITY_TOUCH_TASK,
                   NULL) != pdPASS)
    {
        return(1);
    }

    //
    // Success.
    //
    return(0);
}
