//*****************************************************************************
//
// uart_task.c - Task to output text on the UART.
//
// Copyright (c) 2009-2010 Texas Instruments Incorporated.  All rights reserved.
// Software License Agreement
// 
// Texas Instruments (TI) is supplying this software for use solely and
// exclusively on TI's microcontroller products. The software is owned by
// TI and/or its suppliers, and is protected under applicable copyright
// laws. You may not combine this software with "viral" open-source
// software in order to form a larger program.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND WITH ALL FAULTS.
// NO WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT
// NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. TI SHALL NOT, UNDER ANY
// CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
// DAMAGES, FOR ANY REASON WHATSOEVER.
// 
// This is part of revision 5585 of the EK-LM3S9B96 Firmware Package.
//
//*****************************************************************************

#include "utils/uartstdio.h"
#include "SafeRTOS/SafeRTOS_API.h"
#include "uart_task.h"
#include "priorities.h"

//*****************************************************************************
//
// The item size, queue size, and memory size for the UART message queue.
//
//*****************************************************************************
#define UART_ITEM_SIZE          sizeof(const char *)
#define UART_QUEUE_SIZE         10
#define UART_MEM_SIZE           ((UART_ITEM_SIZE * UART_QUEUE_SIZE) +         \
                                 portQUEUE_OVERHEAD_BYTES)

//*****************************************************************************
//
// A buffer to contain the contents of the UART message queue.
//
//*****************************************************************************
static unsigned long g_pulUARTQueueMem[(UART_MEM_SIZE + 3) / 4];

//*****************************************************************************
//
// The queue that holds messages sent to the UART task.
//
//*****************************************************************************
static xQueueHandle g_pUARTQueue;

//*****************************************************************************
//
// The stack for the UART task.
//
//*****************************************************************************
static unsigned long g_pulUARTTaskStack[128];

//*****************************************************************************
//
// This task receives messages from the other tasks and outputs strings to the
// UART.
//
//*****************************************************************************
static void
UARTTask(void *pvParameters)
{
    const char *pcString;

    //
    // Loop forever.
    //
    while(1)
    {
        //
        // Read the next message from the queue.
        //
        if(xQueueReceive(g_pUARTQueue, &pcString, portMAX_DELAY) == pdPASS)
        {
            //
            // Output this string to the UART.
            //
            UARTprintf(pcString);
        }
    }
}

//*****************************************************************************
//
// Sends a request to the UART task to output a string to the UART.
//
//*****************************************************************************
void
UARTOutputString(const char *pcString)
{
    //
    // Send the string output request to the UART task.
    //
    xQueueSend(g_pUARTQueue, &pcString, portMAX_DELAY);
}

//*****************************************************************************
//
// Initializes the UART task.
//
//*****************************************************************************
unsigned long
UARTTaskInit(void)
{
    //
    // Create a queue for sending messages to the UART task.
    //
    if(xQueueCreate((signed portCHAR *)g_pulUARTQueueMem, UART_MEM_SIZE,
                    UART_QUEUE_SIZE, UART_ITEM_SIZE, &g_pUARTQueue) != pdPASS)
    {
        return(1);
    }

    //
    // Create the UART task.
    //
    if(xTaskCreate(UARTTask, (signed portCHAR *)"UART",
                   (signed portCHAR *)g_pulUARTTaskStack,
                   sizeof(g_pulUARTTaskStack), NULL, PRIORITY_UART_TASK,
                   NULL) != pdPASS)
    {
        return(1);
    }

    //
    // Success.
    //
    return(0);
}
